@@ -52,10 +52,13 @@ Upstream/Server
 from base64 import b64encode
 from contextlib import contextmanager
 
+import errno
 import httplib
 import json
 import pprint
+import time
 import signal
+import socket
 import StringIO
 import logging
 import websocket
@@ -125,6 +128,8 @@ class RPC(object):
     _auth = False
     _request_id = 0
     _debug = False
+    _reconnect_params = None
+    conn = None
 
     def _rpc(self, op):
         if not self._auth and not op.get("Request") == "Login":
@@ -142,12 +147,49 @@ class RPC(object):
             log.debug("rpc response:\n%s" % (json.dumps(result, indent=2)))
 
         if 'Error' in result:
-            # print "raw", op['Request'], raw
             # The backend disconnects us on err, bug: http://pad.lv/1160971
             self.conn.connected = False
             raise EnvError(result)
         return result['Response']
 
+    def login(self, password, user="user-admin", **ignore):
+        """Login gets shared to watchers for reconnect."""
+        if self.conn and self.conn.connected and self._auth:
+            raise AlreadyConnected()
+        # Store for constructing separate authenticated watch connections.
+        self._creds = {'password': password, 'user': user}
+        result = self._rpc(
+            {"Type": "Admin", "Request": "Login",
+             "Params": {"AuthTag": user, "Password": password}})
+        self._auth = True
+        return result
+
+    def set_reconnect_params(self, params):
+        self._reconnect_params = params
+
+    def reconnect(self):
+        if self.conn:
+            self._auth = False
+            self.conn.close()
+        if not self._reconnect_params:
+            return False
+
+        log.info("Reconnecting client")
+        while True:
+            try:
+                self.conn = websocket.create_connection(
+                    self._reconnect_params['url'],
+                    origin=self._reconnect_params['origin'])
+                break
+            except socket.error as err:
+                if not err.errno in (
+                        errno.ETIMEDOUT, errno.ECONNREFUSED, errno.ECONNRESET):
+                    raise
+                time.sleep(1)
+                continue
+        self.login(**self._reconnect_params)
+        return True
+
 
 class Watcher(RPC):
 
@@ -184,9 +226,17 @@ class Watcher(RPC):
             if "state watcher was stopped" in e.message:
                 if not self.auto_reconnect:
                     raise
-                return self._reconnect().next()
+                if not self.reconnect():
+                    raise
+                return self.next()
+            raise
         return result['Deltas']
 
+    def reconnect(self):
+        self.watcher_id = None
+        self.running = False
+        return super(Watcher, self).reconnect()
+
     def stop(self):
         if not self.conn.connected:
             return
@@ -195,6 +245,7 @@ class Watcher(RPC):
             'Request': 'Stop',
             'Id': self.watcher_id})
         self.conn.close()
+        self.watcher_id = None
         self.running = False
         return result
 
@@ -202,12 +253,6 @@ class Watcher(RPC):
         self.context = context
         return self
 
-    def _reconnect(self):
-        self.conn.close()
-        self.conn.connect()
-        self.start()
-        return self
-
     def __iter__(self):
         return self
 
@@ -304,15 +349,6 @@ class Environment(RPC):
              "Params": {"URL": charm_url}})
 
     # Environment operations
-    def login(self, password, user="user-admin"):
-        if self.conn and self.conn.connected and self._auth:
-            raise AlreadyConnected()
-        # Store for constructing separate authenticated watch connections.
-        self._creds = {'password': password, 'user': user}
-        self._rpc({"Type": "Admin", "Request": "Login",
-                   "Params": {"AuthTag": user, "Password": password}})
-        self._auth = True
-
     def info(self):
         return self._rpc({
             "Type": "Client",
@@ -483,6 +519,8 @@ class Environment(RPC):
         else:
             watch_env = connection
 
+        p = dict(self._creds)
+        p.update({'url': self.endpoint, 'origin': self.endpoint})
         if timeout is not None:
             if watch_class is None:
                 watch_class = TimeoutWatcher
@@ -492,6 +530,7 @@ class Environment(RPC):
             if watch_class is None:
                 watch_class = Watcher
             watcher = watch_class(watch_env.conn)
+        watcher.set_reconnect_params(p)
         self._watches.append(watcher)
         watcher.start()
         return watcher
